package cz.vsb.jezek.tetris;
/*
 * Tetris.java
 *
 * Created on 6. 9. 2000, 2:17
 */

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.BufferedInputStream;
import java.io.IOException;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.UnsupportedAudioFileException;

/**
 *
 * @author David Jezek
 * @version
 */
public class Tetris extends java.awt.Component implements GameListener {

	/**
	 * 
	 */
	private static final long serialVersionUID = 3257571693781662008L;

	protected BufferedImage[] square = new BufferedImage[7];
	protected BufferedImage background;
	public Clip music;
	public Clip no;
	public Clip yes;
	private BufferedImage buffer;
	private BufferedImage buffer2;
	private BufferedImage buffer3;
	private TetrisModel model;
	private Rectangle rec = new Rectangle();
	/** Holds value of property scoreListener. */
	private ScoreListener scoreListener;
	private boolean needUpdateB = false;

	/** Initializes the applet Tetris */
	public Tetris(BufferedImage backgroundImage, BufferedImage[] sguares) {
		setBackground(Color.black);
		for (int i = 0; i < sguares.length; i++)
			square[i] = sguares[i];
		background = backgroundImage;
		initSounds();
		addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {
				if (e.getKeyCode() == KeyEvent.VK_LEFT)
					model.moveLeft();
				if (e.getKeyCode() == KeyEvent.VK_RIGHT)
					model.moveRight();
				if (e.getKeyCode() == KeyEvent.VK_UP)
					model.turn();
				if (e.getKeyCode() == KeyEvent.VK_DOWN)
					model.fastDrop();
			}
		});
		addFocusListener(new FocusAdapter() {
			@Override
			public void focusGained(FocusEvent e) {
				model.resume();
			}

			@Override
			public void focusLost(FocusEvent e) {
				model.pauseGame();
			}
		});
		addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent e) {
				requestFocus();
			}
		});
	}

	private void initSounds() {
		try {
			yes = getClip("yes.au");
			no = getClip("no.au");
			music = getClip("music.mid");
		} catch (LineUnavailableException | UnsupportedAudioFileException | IOException e) {
			e.printStackTrace();
		}

	}

	private Clip getClip(String name) throws UnsupportedAudioFileException, IOException, LineUnavailableException {
		Clip clip = AudioSystem.getClip();
		AudioInputStream inputStream = AudioSystem.getAudioInputStream(
				new BufferedInputStream(getClass().getResourceAsStream("/cz/vsb/jezek/tetris/resources/" + name)));
		clip.open(inputStream);
		return clip;
	}

	public void startGame() {
		updateBuffers();
		music.setFramePosition(0);
		music.loop(Clip.LOOP_CONTINUOUSLY);
		requestFocus();
	}

	public void setModel(TetrisModel tm) {
		model = tm;
	}

	public void fallBlockChanged(Block k, int x, int y, Rectangle prev) {
		Graphics g = buffer3.getGraphics();
		int[][] tvar = k.getShape();
		requestFocus();
		prev.y = prev.y < 0 ? 0 : prev.y;
		g.drawImage(buffer2, prev.x, prev.y, prev.x + prev.width, prev.y + prev.height, prev.x, prev.y,
				prev.x + prev.width, prev.y + prev.height, this);
		for (int i = 0; i < tvar.length; i++)
			for (int j = 0; j < tvar[0].length; j++)
				if (tvar[i][j] != 0)
					g.drawImage(square[k.getType() - 1], x + j * model.getSquareSize(),
							y - (tvar.length - i) * model.getSquareSize(), this);
		g = getGraphics();
		rec.setSize(tvar[0].length * model.getSquareSize(), tvar.length * model.getSquareSize());
		rec.setLocation(x, y - rec.height);
		rec.add(prev);
		rec.y = rec.y > 0 ? rec.y : 0;
		if (g != null)
			g.drawImage(buffer3, rec.x, rec.y, rec.x + rec.width, rec.y + rec.height, rec.x, rec.y, rec.x + rec.width,
					rec.y + rec.height, this);
	}

	public void linesDeleted(int[][] area) {
		fillWith(buffer2, background);
		Graphics g = buffer2.getGraphics();
		for (int i = 0; i < area.length; i++)
			for (int j = 0; j < area[0].length; j++)
				if (area[i][j] != 0) {
					g.drawImage(square[area[i][j] - 1], j * model.getSquareSize(), i * model.getSquareSize() - 1, this);
				}
		buffer3.getGraphics().drawImage(buffer2, 0, 0, this);
		yes.setFramePosition(0);
		yes.start();
		paint(getGraphics());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see GameListener#fallKostkaWeld(Kostka, int, int)
	 */
	public void fallBlockWeld(Block k, int x, int y) {
		Graphics g = buffer2.getGraphics();
		int[][] tvar = k.getShape();
		for (int i = 0; i < tvar.length; i++)
			for (int j = 0; j < tvar[0].length; j++)
				if (tvar[i][j] != 0)
					g.drawImage(square[k.getType() - 1], x + j * model.getSquareSize(),
							y - (tvar.length - i) * model.getSquareSize(), this);
		buffer3.getGraphics().drawImage(buffer2, 0, 0, this);
		no.start();
		paint(getGraphics());
	}

	public void initBuffers() {
		buffer = new BufferedImage(model.getSquareSize() * model.getWidth(), model.getDepth() * model.getSquareSize(),
				BufferedImage.TYPE_INT_RGB);
		buffer2 = new BufferedImage(model.getSquareSize() * model.getWidth(), model.getDepth() * model.getSquareSize(),
				BufferedImage.TYPE_INT_RGB);
		buffer3 = new BufferedImage(model.getSquareSize() * model.getWidth(), model.getDepth() * model.getSquareSize(),
				BufferedImage.TYPE_INT_RGB);
		updateBuffers();
		repaint();
	}

	public void updateBuffers() {
		fillWith(buffer, background);
		fillWith(buffer2, background);
		fillWith(buffer3, background);
	}

	private void fillWith(BufferedImage destination, BufferedImage source) {
		for (int i = 0; i < destination.getHeight(); i += source.getHeight()) {
			for (int j = 0; j < destination.getWidth(); j += source.getWidth()) {
				destination.getGraphics().drawImage(source, j, i, this);
			}
		}
	}

	@Override
	public void paint(Graphics g) {
		if (g == null) {
			return;
		}
		if (buffer3 == null) {
			g.setColor(Color.black);
			g.fillRect(0, 0, this.getWidth(), this.getHeight());
		} else {
			if (needUpdateB)
				updateBuffers();
			g.drawImage(buffer3, 0, 0, this);
		}
	}

	@Override
	public Dimension getMinimumSize() {
		return new Dimension(model.getSquareSize() * model.getWidth(), model.getSquareSize() * model.getDepth());
	}

	@Override
	public Dimension getPreferredSize() {
		return getMinimumSize();
	}

	public void setScoreListener(ScoreListener scoreListener) {
		this.scoreListener = scoreListener;
	}

	public void endGame() {
		music.stop();
	}

	public void pauseGame() {
		music.stop();
	}

	public void resumeGame() {
		music.loop(Clip.LOOP_CONTINUOUSLY);
	}
}