package cz.vsb.jezek.tetris;
/*
 * TetrisModel.java
 *
 * Created on 6. z��� 2000, 22:44
 */

import java.awt.Rectangle;
import java.util.Random;

/**
 *
 * @author David Jezek
 * @version 1.0
 */
public class TetrisModel implements Runnable {

	public static final int GAME_RUNNING = 0;
	public static final int GAME_PAUSED = 1;
	public static final int GAME_OFF = 2;

	private int depthOfPlayArea;
	private int arerWidth;
	private int squareSize;

	private int[][] area;
	private Block actualBlock = new Block();
	private int level;
	private int speed;
	private int gameState;
	private int blockAreaPositionX;
	private int blockAreaPositionY;
	private int blockPixelPositionX;
	private int blockPixelPositionY;
	private boolean endOfGame;
	private int movePixelStep;
	private Random random;
	private Rectangle previous = new Rectangle();
	private boolean fastDropMode = false;
	private boolean doPause = false;
	/** Holds value of property score. */
	private long score;
	/** Holds value of property gameListener. */
	private GameListener gameListener;
	/** Holds value of property scoreListener. */
	private ScoreListener scoreListener;

	/** Creates new TetrisModel */
	public TetrisModel() {
		this(20, 10, 20);
	}

	public TetrisModel(int depth, int width, int squareSize) {
		this.depthOfPlayArea = depth;
		this.arerWidth = width;
		this.squareSize = squareSize;
		gameState = GAME_OFF;
		random = new Random();
		area = new int[depth][width];
	}

	public void initNewGame() {
		if (gameState == GAME_RUNNING) {
			waitToGameStateInitToNothing();
		}
		for (int i = 0; i < area.length; i++) {
			for (int j = 0; j < area[0].length; j++) {
				area[i][j] = 0;
			}
		}
		setScore(0);
		setLevel(1);
		gameState = GAME_RUNNING;
		endOfGame = false;
		sentEventLineDeleted();
	}

	private void sentEventLineDeleted() {
		if (gameListener != null) {
			gameListener.linesDeleted(area);
		}
	}

	private void waitToGameStateInitToNothing() {
		synchronized (this) {
			while (gameState != GAME_OFF && !Thread.currentThread().isInterrupted()) {
				try {
					wait();
				} catch (InterruptedException ex) {
					Thread.currentThread().interrupt();
				}
			}
		}
	}

	public void newBlock() {
		blockAreaPositionX = arerWidth / 2 - 1;
		blockAreaPositionY = 0;
		blockPixelPositionX = blockAreaPositionX * squareSize;
		blockPixelPositionY = (blockAreaPositionY + 1) * squareSize - 1;
		movePixelStep = 1;
		speed = level * 5 + 20; // speed in picxel peer second
		actualBlock.setType(random.nextInt(7)+1);
		actualBlock.setOrientation(Block.ANGLE_0);
		fastDropMode = false;
	}

	public void run() {
		sendEventStartGame();
		initNewGame();
		newBlock();
		endOfGame = false;
		while (!endOfGame && !Thread.currentThread().isInterrupted()) {
			long sleepTime = movePixelStep * 1000 / speed;
			for (; sleepTime < 20; sleepTime = movePixelStep * 1000 / speed) {
				movePixelStep++;
			}
			sleep(sleepTime);
			if (doPause) {
				pauseGameWait();
			}
			MoveDownResult moveDownResult = moveDownAndTestTouche(); 
			if (!moveDownResult.touche) {
				moveDown(moveDownResult.moveDownsizeInAreaGrid);
			} else {
				weld();
				newBlock();
			}
			if (!endOfGame) {
				endOfGame = testEnd();
			}
		}
		synchronized (this) {
			gameState = GAME_OFF;
			this.notifyAll();
		}
		sentEventGameEnd();
	}

	private void sentEventGameEnd() {
		if (gameListener != null) {
			gameListener.endGame();
		}
	}

	private void sendEventStartGame() {
		if (gameListener != null) {
			gameListener.startGame();
		}
	}

	private void sleep(long sleepTime) {
		try {
			Thread.sleep(sleepTime);
		} catch (InterruptedException ex) {
			Thread.currentThread().interrupt();
		}
	}

	private void pauseGameWait() {
		synchronized (this) {
			gameState = GAME_PAUSED;
			gameListener.pauseGame();
			try {
				this.wait();
			} catch (InterruptedException ex) {
				Thread.currentThread().interrupt();
			}
			doPause = false;
			gameState = GAME_RUNNING;
			gameListener.resumeGame();
		}
	}

	protected synchronized MoveDownResult moveDownAndTestTouche() {
		int posibleMovementSizeInAreaGrid = 0;
		int movementSizeInAreaGrid = getBlockYPositionInAreaGrid(blockPixelPositionY + movePixelStep)
				- getBlockYPositionInAreaGrid(blockPixelPositionY);
		if (movementSizeInAreaGrid == 0) {
			return new MoveDownResult(0, false);
		}
		int[][] blockCubes = actualBlock.getShape();
		posibleMovementSizeInAreaGrid = movementSizeInAreaGrid;
		for (int l = 1; l <= movementSizeInAreaGrid; l++) {
			MoveDownResult moveDownResult = testDownMoveByGridArea(blockCubes, l);
			if(moveDownResult.touche) {
				return new MoveDownResult(posibleMovementSizeInAreaGrid, true);
			}
		}
		if (fastDropMode)
			setScore(getScore() + posibleMovementSizeInAreaGrid);
		return new MoveDownResult(movementSizeInAreaGrid, false);
	}
	
	private MoveDownResult testDownMoveByGridArea(int[][] blockCubes, int moveLengthInGridArea) {
		for (int i = 0; i < blockCubes.length; i++) {
			for (int j = 0; j < blockCubes[0].length; j++) {
				if (!isCubeOverPlayArea(blockCubes, blockAreaPositionY + moveLengthInGridArea, i)) {
					if (cubeYPositionInGameArea(blockCubes, blockAreaPositionY + moveLengthInGridArea, i) > depthOfPlayArea-1
							|| !(blockCubes[i][j] == 0 || area[cubeYPositionInGameArea(blockCubes, blockAreaPositionY + moveLengthInGridArea, i)][j
									+ blockAreaPositionX] == 0)) {
						int resultedMoveLengthInGridArea = moveLengthInGridArea - 1;
						blockAreaPositionY = blockAreaPositionY + resultedMoveLengthInGridArea;
						blockPixelPositionY = (blockAreaPositionY + 1) * squareSize - 1;
						if (fastDropMode && (moveLengthInGridArea - 1 != 0)) {
							setScore(getScore() + resultedMoveLengthInGridArea);
						}
						return new MoveDownResult(resultedMoveLengthInGridArea, true);
					}
				}
			}
		}
		return new MoveDownResult(moveLengthInGridArea, false);
	}

	private int getBlockYPositionInAreaGrid(int yPositionInPixel) {
		return yPositionInPixel / squareSize;
	}

	protected synchronized void moveDown(int moveDownsizeInAreaGrid) {
		if (getBlockYPositionInAreaGrid(blockPixelPositionY) != getBlockYPositionInAreaGrid(
				blockPixelPositionY + movePixelStep)) {
			blockAreaPositionY += moveDownsizeInAreaGrid;
		}
		int[][] tvar = actualBlock.getShape();
		previous.setSize(tvar[0].length * squareSize, tvar.length * squareSize);
		previous.setLocation(blockPixelPositionX, blockPixelPositionY - previous.height);
		blockPixelPositionY += movePixelStep;
		sentEventFallBlockChanged();
	}

	private void sentEventFallBlockChanged() {
		if (gameListener != null) {
			gameListener.fallBlockChanged(actualBlock, blockPixelPositionX, blockPixelPositionY, previous);
		}
	}

	protected synchronized void weld() {
		int[][] blockCubes = actualBlock.getShape();
		boolean[] linesToDelete = new boolean[depthOfPlayArea];
		blockPixelPositionY = (blockAreaPositionY + 1) * squareSize - 1;
		blockPixelPositionX = blockAreaPositionX * squareSize;
		weldCubes(blockCubes);
		findLineToDelete(blockCubes, linesToDelete);
		int deletedLineCount = deleteLines(linesToDelete);
		if (deletedLineCount > 0) {
			sentEventLinesDeleted();
			int addScore = 0;
			switch (deletedLineCount) {
			case 1:
				addScore = 10;
				break;
			case 2:
				addScore = 30;
				break;
			case 3:
				addScore = 50;
				break;
			case 4:
				addScore = 80;
				break;
			default:
				addScore = deletedLineCount * 10;
				break;
			}
			setScore(getScore() + addScore);
		} else {
			sentEventFallBlockWeld();
		}
	}

	private int deleteLines(boolean[] linesToDelete) {
		int j;
		int i;
		int deletedLineCount = 0;
		for (i = area.length - 1, j = area.length - 1; j >= 0;) {
			if (linesToDelete[j]) {
				deletedLineCount++;
				j--;
				continue;
			}
			area[i] = area[j];
			i--;
			j--;
		}
		for (; i >= 0; i--) {
			area[i] = new int[arerWidth];
		}
		return deletedLineCount;
	}

	private void findLineToDelete(int[][] blockCubes, boolean[] liensTodelete) {
		boolean deleteLine = false;
		for (int i = blockCubes.length - 1; i >= 0; i--) {
			deleteLine = true;
			for (int j = area[0].length - 1; j >= 0; j--) {
				if (blockAreaPositionY - i >= 0 && area[blockAreaPositionY - i][j] == 0) {
					deleteLine = false;
					break;
				}
			}
			if (deleteLine && blockAreaPositionY - i >= 0) {
				liensTodelete[blockAreaPositionY - i] = true;
			}
		}
	}

	private void weldCubes(int[][] blockCubes) {
		for (int i = 0; i < blockCubes.length; i++) {
			for (int j = 0; j < blockCubes[0].length; j++) {
				if (!isCubeOverPlayArea(blockCubes, i)) {
					area[i + blockAreaPositionY - blockCubes.length + 1][j + blockAreaPositionX] += blockCubes[i][j];
				}
			}
		}
	}

	protected synchronized boolean testEnd() {
		for (int i = 0; i < area[0].length; i++) {
			if (area[0][i] != 0) {
				return true;
			}
		}
		return false;
	}

	protected synchronized boolean moveRight() {
		int[][] blockCubes = actualBlock.getShape();
		int newLeftCornerX = blockAreaPositionX + 1;
		for (int i = 0; i < blockCubes.length; i++) {
			for (int j = 0; j < blockCubes[0].length; j++) {
				if (newLeftCornerX + j >= area[0].length) {
					return false;
				}
				if (isCubeOverPlayArea(blockCubes, i)) {
					continue;
				}
				if (!(isInsidePlayArea(blockCubes, newLeftCornerX, blockAreaPositionY, i, j)
						&& isNoCollisionOfCubes(blockCubes, newLeftCornerX, blockAreaPositionY, i, j))) {
					return false;
				}
			}
		}
		if (blockYPixelOffsetToAreaGrid() < squareSize/2) {
			for (int i = 0; i < blockCubes.length; i++) {
				for (int j = 0; j < blockCubes[0].length; j++) {
					if (isCubeOverPlayArea(blockCubes, blockAreaPositionY - 1, i)) {
						continue;
					}
					if (!(isInsidePlayArea(blockCubes, newLeftCornerX, blockAreaPositionY-1, i, j)
							&& isNoCollisionOfCubes(blockCubes, newLeftCornerX, blockAreaPositionY - 1, i, j)))
						return false;
				}
			}
		}
		previous.setSize(blockCubes[0].length * squareSize, blockCubes.length * squareSize);
		previous.setLocation(blockPixelPositionX, blockPixelPositionY - previous.height);
		blockAreaPositionX++;
		blockPixelPositionX = blockAreaPositionX * squareSize;
		sentEventFallBlockChanged();
		return true;
	}

	private int blockYPixelOffsetToAreaGrid() {
		return blockPixelPositionY % squareSize;
	}

	protected synchronized boolean moveLeft() {
		int[][] blockCubes = actualBlock.getShape();
		int newLeftCornerX = blockAreaPositionX - 1;
		for (int i = 0; i < blockCubes.length; i++) {
			for (int j = 0; j < blockCubes[0].length; j++) {
				if (newLeftCornerX + j < 0) {
					return false;
				}
				if (isCubeOverPlayArea(blockCubes, i)) {
					continue;
				}
				if (!(isInsidePlayArea(blockCubes, newLeftCornerX, blockAreaPositionY, i, j)
						&& isNoCollisionOfCubes(blockCubes, newLeftCornerX, blockAreaPositionY, i, j))) {
					return false;
				}
			}
		}
		if (blockYPixelOffsetToAreaGrid() < squareSize/2) {
			for (int i = 0; i < blockCubes.length; i++) {
				for (int j = 0; j < blockCubes[0].length; j++) {
					if (isCubeOverPlayArea(blockCubes, i)) {
						continue;
					}
					if (!(isInsidePlayArea(blockCubes, newLeftCornerX, blockAreaPositionY, i, j)
							&& isNoCollisionOfCubes(blockCubes, newLeftCornerX, blockAreaPositionY - 1, i, j))) {
						return false;
					}
				}
			}
		}
		previous.setSize(blockCubes[0].length * squareSize, blockCubes.length * squareSize);
		previous.setLocation(blockPixelPositionX, blockPixelPositionY - previous.height);
		blockAreaPositionX--;
		blockPixelPositionX = blockAreaPositionX * squareSize;
		sentEventFallBlockChanged();
		return true;
	}

	protected synchronized boolean turn() {
		int[][] blockCubes = actualBlock.getShape();
		previous.setSize(blockCubes[0].length * squareSize, blockCubes.length * squareSize);
		previous.setLocation(blockPixelPositionX, blockPixelPositionY - previous.height);
		int blockSizeMiddleY = blockCubes.length / 2;
		int blockSizeMiddleX = blockCubes[0].length / 2;
		actualBlock.setOrientation((actualBlock.getOrientation() + 1) % 4);
		int[][] newBlockCubes = actualBlock.getShape();
		int newLeftUpperCornerX = blockAreaPositionX + blockSizeMiddleX - newBlockCubes[0].length / 2;
		int newLeftUpperCornerY = blockAreaPositionY + newBlockCubes.length / 2 - blockSizeMiddleY;

		for (int i = 0; i < newBlockCubes.length; i++) {
			for (int j = 0; j < newBlockCubes[0].length; j++) {
				if (isCubeOverPlayArea(newBlockCubes, newLeftUpperCornerY, i)) {
					continue;
				}
				if (isInsidePlayArea(newBlockCubes, newLeftUpperCornerX, newLeftUpperCornerY, i, j)
						&& isNoCollisionOfCubes(newBlockCubes, newLeftUpperCornerX, newLeftUpperCornerY, i, j)) {
					continue;
				}
				actualBlock.setOrientation((actualBlock.getOrientation() - 1 + 4) % 4);
				return false;
			}
		}
		blockAreaPositionX = newLeftUpperCornerX;
		blockPixelPositionY = blockPixelPositionY + (newLeftUpperCornerY - blockAreaPositionY) * squareSize;
		blockAreaPositionY = newLeftUpperCornerY;
		blockPixelPositionX = blockAreaPositionX * squareSize;
		sentEventFallBlockChanged();
		return true;
	}

	private boolean isCubeOverPlayArea(int[][] blockCubes, int i) {
		return isCubeOverPlayArea(blockCubes, blockAreaPositionY, i);
	}

	private boolean isCubeOverPlayArea(int[][] blockCubes, int leftUpperCornerY, int i) {
		return cubeYPositionInGameArea(blockCubes, leftUpperCornerY, i) < 0;
	}

	private int cubeYPositionInGameArea(int[][] blockCubes, int leftUpperCornerY, int i) {
		return i + leftUpperCornerY - blockCubes.length + 1;
	}

	private boolean isNoCollisionOfCubes(int[][] blockCubes, int leftUpperCornerX, int leftUpperCornerY, int i, int j) {
			if (!isCubeOverPlayArea(blockCubes, leftUpperCornerY, i)) {
				return blockCubes[i][j] == 0
						|| area[cubeYPositionInGameArea(blockCubes, i, leftUpperCornerY)][leftUpperCornerX + j] == 0;
			}
		return true;
	}

	private boolean isInsidePlayArea(int[][] blockCubes, int leftUpperCornerX, int leftUpperCornerY, int i, int j) {
		return leftUpperCornerX + j >= 0 && leftUpperCornerX + j < area[0].length
				&& cubeYPositionInGameArea(blockCubes, i, leftUpperCornerY) >= 0
				&& cubeYPositionInGameArea(blockCubes, i, leftUpperCornerY) < area.length;
	}

	protected synchronized boolean fastDrop() {
		speed = 1000;
		fastDropMode = true;
		return false;
	}

	/**
	 * Getter for property level.
	 * 
	 * @return Value of property level.
	 */
	public int getLevel() {
		return level;
	}

	/**
	 * Setter for property level.
	 * 
	 * @param level New value of property level.
	 */
	public void setLevel(int level) {
		this.level = level;
		if (scoreListener != null)
			scoreListener.levelChanged(level);
	}

	/**
	 * Getter for property score.
	 * 
	 * @return Value of property score.
	 */
	public long getScore() {
		return score;
	}

	/**
	 * Setter for property score.
	 * 
	 * @param score New value of property score.
	 *
	 */
	public void setScore(long score) {
		if (this.score / 500 != score / 500)
			setLevel(getLevel() + 1);
		this.score = score;
		if (scoreListener != null)
			scoreListener.scoreChanged(score);
	}

	/**
	 * Setter for property gameListener.
	 * 
	 * @param gameListener New value of property gameListener.
	 */
	public void setGameListener(GameListener gameListener) {
		this.gameListener = gameListener;
		gameListener.setModel(this);
	}

	/**
	 * Setter for property scoreListener.
	 * 
	 * @param scoreListener New value of property scoreListener.
	 */
	public void setScoreListener(ScoreListener scoreListener) {
		if (scoreListener != null) {
			this.scoreListener = scoreListener;
		}
	}

	public void endGame() {
		if (gameState == GAME_RUNNING || gameState == GAME_PAUSED) {
			endOfGame = true;
		}
	}

	public void pauseGame() {
		if (gameState == GAME_RUNNING) {
			doPause = true;
		}
	}

	public void resume() {
		if (gameState == GAME_PAUSED) {
			synchronized (this) {
				notifyAll();
			}
		}
	}

	public int getDepth() {
		return depthOfPlayArea;
	}

	public int getWidth() {
		return arerWidth;
	}

	private void sentEventFallBlockWeld() {
		if (gameListener != null)
			gameListener.fallBlockWeld(actualBlock, blockPixelPositionX, blockPixelPositionY);
	}

	private void sentEventLinesDeleted() {
		if (gameListener != null) {
			gameListener.linesDeleted(area);
		}
	}

	public int getSquareSize() {
		return squareSize;
	}
	
	private class MoveDownResult{
		public final int moveDownsizeInAreaGrid;
		public final boolean touche;
		public MoveDownResult(int moveDownsizeInAreaGrid, boolean touche) {
			super();
			this.moveDownsizeInAreaGrid = moveDownsizeInAreaGrid;
			this.touche = touche;
		}
	}
	

}